"""Widget to select a range of float values"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
from ipywidgets import widgets
import ipyvuetify as v

try:
    from . import settings
    from . import label
    from . import tooltip
except:
    import settings
    import label
    import tooltip


#####################################################################################################################################################
# rangeSliderFloat control
#####################################################################################################################################################
class rangeSliderFloat():
    """
    Compound widget to select a range of float values in a specific range.
        
    Parameters
    ----------
    selectedminvalue : float
        Initial minimum value of the slider
    selectedmaxvalue : float
        Initial maximum value of the slider
    minvalue : float, optional
        Minimum value of the slider (default is 0.0)
    maxvalue : float, optional
        Maximum value of the slider (default is 1.0)
    text : str, optional
        Text to display on the left of the slider (default is 'Select')
    showpercentage : bool, optional
        If True, the widget will write the current value as a percentage inside the allowed range and will append the % sign to the right of the current value (default is True)
    decimals : int, optional
        Number of decimal digits to use in case showpercentage is False (default is 2)
    maxint : int, optional
        Maximum integer number for the underlining integer slider (defines the slider sensitivity). Default value is None, meaning it will be automatically calculated
    labelwidth : int, optional
        Width of the label part of the widgets in pixels (default is 0, meaning it is automatically calculated based on the provided label text)
    sliderwidth : int, optional
        Width in pixels of the slider component of the widget (default is 200)
    resetbutton: bool, optional
        If True a reset button is displayed, allowing for resetting the widget to its initial value (default is False)
    showtooltip: bool, optional
        If True the up and down buttons will have a tooltip (default is False)
    onchange : function, optional
        Python function to call when the changes the range value of the slider. The function will receive a single parameter, containing the new value of the opacity in the range from 0.0 to 1.0 (default is None)
    editable : bool, optional
        If True the min and max label can be edited by clicking on them (default is False)
    editableWidth : int, optional
        If the slider is editable, set the width of the v.TextField widgets to enter the min and max value (default is 90)
            
    Example
    -------
    Creation and display of a range slider widget to select an opacity range in a custom interval::
        
        from vois.vuetify import rangeSliderFloat
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def onchange(value):
            with output:
                print(value)

        o = rangeSliderFloat.rangeSliderFloat(1.8, 2.5, text='Select Value in [1.0,3.0]:',
                                              minvalue=1.0, maxvalue=3.0, onchange=onchange)
        
        display(o.draw())
        display(output)

    .. figure:: figures/rangeSliderFloat.png
       :scale: 100 %
       :alt: opacity widget

       Example of an range slider widget to select a floating point range.
   """

    # Initialization
    def __init__(self, selectedminvalue, selectedmaxvalue, minvalue=0.0, maxvalue=1.0, text='Select', showpercentage=True, decimals=2, maxint=None, 
                 labelwidth=0, sliderwidth=200, resetbutton=False, showtooltip=False, onchange=None,
                 editable=False, editableWidth=90):
        
        self.onchange = onchange
        
        self.editable = editable
        self.editableWidth = editableWidth
        
        self.minvalue = minvalue
        self.maxvalue = maxvalue
        self.showpercentage = showpercentage
        self.decimals       = decimals
        
        if maxint is None:
            if self.decimals <= 1:
                self.maxint = 10
            elif self.decimals == 2:
                self.maxint = 100
            elif self.decimals == 3:
                self.maxint = 1000
            elif self.decimals == 4:
                self.maxint = 10000
            else:
                self.maxint = 100000
        else:
            self.maxint = maxint
        
        self.postchar = ''
        if self.showpercentage: self.postchar = '%'
            
        self.labelwidth  = labelwidth
        self.sliderwidth = sliderwidth
        self.resetbutton = resetbutton
        
        self.intvaluemin = self.float2integer(selectedminvalue)
        self.intvaluemax = self.float2integer(selectedmaxvalue)
        
        #self.label  = label.label(text, textweight=400, margins=0, margintop=4, height=22)
        
        if self.labelwidth > 0:
            spx = '%dpx'%self.labelwidth
            style = 'width: %s; min-width: %s; max-width: %s;'%(spx,spx,spx)
        else:
            style = ''
        self.label = v.Html(tag='div', children=[text], class_='pa-0 ma-0 mt-4', style_=style)
        self.slider = v.RangeSlider(v_model=[self.intvaluemin,self.intvaluemax],
                                    dense=True, xsmall=True, 
                                    ticks=False, thumb_size=10, dark=settings.dark_mode,
                                    color=settings.color_first, track_color="grey",
                                    class_="pa-0 ma-0 ml-3 mr-5 mt-3 mb-n1",
                                    style_='max-width: %dpx; width: %dpx;'%(self.sliderwidth,self.sliderwidth),
                                    min=0, max=self.maxint, vertical=False, height=32)
        self.slider.on_event('input',  self.oninput)
        self.slider.on_event('change', self.onsliderchange)

        self.fieldmin = v.TextField(autofocus=True, hide_details=True, single_line=True, hide_spin_buttons=False, dense=True, outlined=True, color=settings.color_first, type="number", class_='pa-0 ma-0 mt-2')
        self.cfieldmin = v.Card(flat=True, children=[self.fieldmin], width=self.editableWidth, max_width=self.editableWidth)
        self.fieldmin.on_event('change', self.onchangefieldmin)
        self.fieldmin.on_event('blur',   self.onchangefieldmin)
        self.fieldmax = v.TextField(autofocus=True, hide_details=True, single_line=True, hide_spin_buttons=False, dense=True, outlined=True, color=settings.color_first, type="number",class_='pa-0 ma-0 mt-2')
        self.cfieldmax = v.Card(flat=True, children=[self.fieldmax], width=self.editableWidth, max_width=self.editableWidth)
        self.fieldmax.on_event('change', self.onchangefieldmax)
        self.fieldmax.on_event('blur',   self.onchangefieldmax)
        
        if self.showpercentage:
            self.labelvaluemin = v.Html(tag='div', children=[str(self.intvaluemin) + self.postchar], class_='pa-0 ma-0 mt-4 ml-4')
            self.labelvaluemax = v.Html(tag='div', children=[str(self.intvaluemax) + self.postchar], class_='pa-0 ma-0 mt-4')
            self.fieldmin.v_model = self.intvaluemin
            self.fieldmax.v_model = self.intvaluemax
        else:
            self.labelvaluemin = v.Html(tag='div', children=['{:.{prec}f}'.format(selectedminvalue, prec=self.decimals) + self.postchar], class_='pa-0 ma-0 mt-4 ml-4')
            self.labelvaluemax = v.Html(tag='div', children=['{:.{prec}f}'.format(selectedmaxvalue, prec=self.decimals) + self.postchar], class_='pa-0 ma-0 mt-4')
            self.fieldmin.v_model = selectedminvalue
            self.fieldmax.v_model = selectedmaxvalue

        self.labelvaluemin.on_event('click', self.onvaluemin)
        self.labelvaluemax.on_event('click', self.onvaluemax)
       
        
        if self.resetbutton:
            self.bupmin = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-menu-right'])])
            self.bupmin.on_event('click', self.onupmin)
            self.cupmin = v.Card(children=[self.bupmin], elevation=0, class_='pa-0 ma-0 mr-1 mt-4', style_="overflow: hidden;")
            if showtooltip: self.cupmin = tooltip.tooltip("Increase min",self.cupmin)

            self.bdnmin = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-menu-left'])])
            self.bdnmin.on_event('click', self.ondnmin)
            self.cdnmin = v.Card(children=[self.bdnmin], elevation=0, class_='pa-0 ma-0 mr-1 mt-4', style_="overflow: hidden;")
            if showtooltip: self.cdnmin = tooltip.tooltip("Decrease min",self.cdnmin)

            self.bresmin = v.Btn(icon=True, x_small=True, rounded=False, elevation=0, width=15, height=15, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-close'])])
            self.bresmin.on_event('click', self.onresetmin)
            self.cresmin = v.Card(children=[self.bresmin], elevation=0, class_='pa-0 ma-0 mr-1 mt-4', style_="overflow: hidden;")
            if showtooltip: self.cresmin = tooltip.tooltip("Reset min value",self.cresmin)
                
            self.buttonsmin = widgets.HBox([self.cdnmin,self.cresmin,self.cupmin])

            self.bupmax = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-menu-right'])])
            self.bupmax.on_event('click', self.onupmax)
            self.cupmax = v.Card(children=[self.bupmax], elevation=0, class_='pa-0 ma-0 mr-1 mt-4', style_="overflow: hidden;")
            if showtooltip: self.cupmax = tooltip.tooltip("Increase max",self.cupmax)

            self.bdnmax = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-menu-left'])])
            self.bdnmax.on_event('click', self.ondnmax)
            self.cdnmax = v.Card(children=[self.bdnmax], elevation=0, class_='pa-0 ma-0 mr-1 mt-4', style_="overflow: hidden;")
            if showtooltip: self.cdnmax = tooltip.tooltip("Decrease max",self.cdnmax)

            self.bresmax = v.Btn(icon=True, x_small=True, rounded=False, elevation=0, width=15, height=15, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-close'])])
            self.bresmax.on_event('click', self.onresetmax)
            self.cresmax = v.Card(children=[self.bresmax], elevation=0, class_='pa-0 ma-0 mr-1 mt-4', style_="overflow: hidden;")
            if showtooltip: self.cresmax = tooltip.tooltip("Reset max value",self.cresmax)
                
            self.buttonsmax = widgets.HBox([self.cdnmax,self.cresmax,self.cupmax])
        else:
            self.bupmin = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-menu-up'])])
            self.bupmin.on_event('click', self.onupmin)
            self.cupmin = v.Card(children=[self.bupmin], elevation=0, class_='pa-0 ma-0 mr-1 mt-2', style_="overflow: hidden;")
            if showtooltip: self.cupmin = tooltip.tooltip("Increase min",self.cupmin)

            self.bdnmin = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-menu-down'])])
            self.bdnmin.on_event('click', self.ondnmin)
            self.cdnmin = v.Card(children=[self.bdnmin], elevation=0, class_='pa-0 ma-0 mr-1 mt-n1', style_="overflow: hidden;")
            if showtooltip: self.cdnmin = tooltip.tooltip("Decrease min",self.cdnmin)

            self.buttonsmin = widgets.VBox([self.cupmin,self.cdnmin])

            self.bupmax = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-menu-up'])])
            self.bupmax.on_event('click', self.onupmax)
            self.cupmax = v.Card(children=[self.bupmax], elevation=0, class_='pa-0 ma-0 mr-1 mt-2', style_="overflow: hidden;")
            if showtooltip: self.cupmax = tooltip.tooltip("Increase max",self.cupmax)

            self.bdnmax = v.Btn(icon=True, small=True, rounded=False, elevation=0, width=15, height=20, class_='pa-0 ma-0', children=[v.Icon(color='grey', children=['mdi-menu-down'])])
            self.bdnmax.on_event('click', self.ondnmax)
            self.cdnmax = v.Card(children=[self.bdnmax], elevation=0, class_='pa-0 ma-0 mr-1 mt-n1', style_="overflow: hidden;")
            if showtooltip: self.cdnmax = tooltip.tooltip("Decrease max",self.cdnmax)

            self.buttonsmax = widgets.VBox([self.cupmax,self.cdnmax])
        
        self.row = v.Row(justify='start', class_='pa-0 ma-0', no_gutters=True, children=[self.label, self.labelvaluemin, self.buttonsmin, self.slider, self.buttonsmax, self.labelvaluemax], style_="overflow: hidden;")
        
        
    # When carriage return on the self.fieldmin or an external click
    def onchangefieldmin(self, *args):
        self.row.children = [self.label, self.labelvaluemin, self.buttonsmin, self.slider, self.buttonsmax, self.labelvaluemax]
        v = self.value
        self.value = float(self.fieldmin.v_model), v[1]
        self.fieldmin.v_model = self.value[0]
        
    # When carriage return on the self.fieldmax or an external click
    def onchangefieldmax(self, *args):
        self.row.children = [self.label, self.labelvaluemin, self.buttonsmin, self.slider, self.buttonsmax, self.labelvaluemax]
        v = self.value
        self.value = v[0], float(self.fieldmax.v_model)
        self.fieldmax.v_model = self.value[1]
        
    # On click on the self.labelvaluemin: display the self.fieldmin
    def onvaluemin(self, *args):
        if self.editable:
            self.fieldmin.v_model = self.value[0]
            self.row.children = [self.label, self.cfieldmin, self.buttonsmin, self.slider, self.buttonsmax, self.labelvaluemax]
        
    # On click on the self.labelvaluemax: display the self.fieldmax
    def onvaluemax(self, *args):
        if self.editable:
            self.fieldmax.v_model = self.value[1]
            self.row.children = [self.label, self.labelvaluemin, self.buttonsmin, self.slider, self.buttonsmax, self.cfieldmax]
        
        
    # Input event on the slider
    def oninput(self, *args):
        if self.showpercentage:
            valuemin,valuemax = self.slider.v_model
            self.labelvaluemin.children=[str(valuemin) + self.postchar]
            self.labelvaluemax.children=[str(valuemax) + self.postchar]
        else:
            valuemin,valuemax = self.slider.v_model
            valuemin = self.integer2float(valuemin)
            valuemax = self.integer2float(valuemax)
            self.labelvaluemin.children=['{:.{prec}f}'.format(valuemin, prec=self.decimals) + self.postchar]
            self.labelvaluemax.children=['{:.{prec}f}'.format(valuemax, prec=self.decimals) + self.postchar]

            
    # Input change on the slider
    def onsliderchange(self, *args):
        if self.showpercentage:
            valuemin,valuemax = self.slider.v_model
            self.labelvaluemin.children=[str(valuemin) + self.postchar]
            self.labelvaluemax.children=[str(valuemax) + self.postchar]
        else:
            valuemin,valuemax = self.slider.v_model
            valuemin = self.integer2float(valuemin)
            valuemax = self.integer2float(valuemax)
            self.labelvaluemin.children=['{:.{prec}f}'.format(valuemin, prec=self.decimals) + self.postchar]
            self.labelvaluemax.children=['{:.{prec}f}'.format(valuemax, prec=self.decimals) + self.postchar]
        
        if not self.onchange is None:
            self.onchange([self.integer2float(x) for x in self.slider.v_model])
        v1,v2 = self.slider.v_model
        self.bupmin.disabled = v1 >= v2
        self.bdnmin.disabled = v1 <= 0
        self.bupmax.disabled = v2 >= self.maxint
        self.bdnmax.disabled = v2 <= v1
            
            
    # Click on the up button
    def onupmin(self, *args):
        v1,v2 = self.slider.v_model
        if v1 < v2:
            self.slider.v_model = [v1+1, v2]
            self.onsliderchange()
    
    # Click on the down button
    def ondnmin(self, *args):
        v1,v2 = self.slider.v_model
        if v1 > 0:
            self.slider.v_model = [v1-1, v2]
            self.onsliderchange()
    
    # Click on the reset min button
    def onresetmin(self, *args):
        v1,v2 = self.slider.v_model
        self.slider.v_model = [self.intvaluemin, v2]
        self.onsliderchange()

        
    # Click on the up button
    def onupmax(self, *args):
        v1,v2 = self.slider.v_model
        if v2 < self.maxint:
            self.slider.v_model = [v1, v2+1]
            self.onsliderchange()
    
    # Click on the down button
    def ondnmax(self, *args):
        v1,v2 = self.slider.v_model
        if v2 > v1:
            self.slider.v_model = [v1, v2-1]
            self.onsliderchange()
        
    # Click on the reset max button
    def onresetmax(self, *args):
        v1,v2 = self.slider.v_model
        self.slider.v_model = [v1, self.intvaluemax]
        self.onsliderchange()
        
        
    # Conversion of float value [minvalue,maxvalue] to integer in [0,self.maxint]
    def float2integer(self, value):
        return int(self.maxint * (max(self.minvalue, min(self.maxvalue, float(value))) - self.minvalue) / (self.maxvalue - self.minvalue))
    
    # Conversion of integer value [0,self.maxint] to float in [0.0,1.0]
    def integer2float(self, value):
        return ((self.maxvalue - self.minvalue) * float(value)) / float(self.maxint) + self.minvalue

    # Draw the widget
    def draw(self):
        """Returns the ipyvuetify object to display (a v.Row widget)"""
        return self.row

        
    # Get the slider value
    @property
    def value(self):
        """
        Get/Set the slider range value as a list of min and max selected value.
        
        Returns
        --------
        value : list of 2 floats
            Current value of the range slider

        Example
        -------
        Programmatically set the slider value
            
            s.value = [0.56, 0.83]
        
        """
        return [self.integer2float(x) for x in self.slider.v_model]
   

    # Set the slider value
    @value.setter
    def value(self, newvalue):
        v1,v2 = newvalue

        if v1 > v2: v1,v2 = v2,v1
        if v1 < self.minvalue:
            v1 = self.minvalue
        if v2 > self.maxvalue:
            v2 = self.maxvalue
            
        intvaluemin = self.float2integer(v1)
        intvaluemax = self.float2integer(v2)
        self.slider.v_model = [intvaluemin,intvaluemax]
        self.onsliderchange()



"""Fullscreen page"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Imports
import ipyvuetify as v

try:
    from . import settings
    from . import dialogGeneric
    from . import tooltip
except:
    import settings
    import dialogGeneric
    import tooltip


#####################################################################################################################################################
# Open a page in fullscreen mode
#####################################################################################################################################################
class page():
    """
    Fullsceen page with title and footer bar.
        
    Parameters
    ----------
    appname : str
        Name of the application. It will be displayed on the left side of the title bar
    title : str
        Title of the page
    output : instance of widgets.Output() class
        Output widget to be used for the opening of the fullscreen dialog that implements the page
    onclose : function, optional
        Python function to call when the user closes the page. The function will receive no parameters (default is None)
    titlecolor : str, optional
        Color to use for the title bar background (default is settings.color_first)
    titledark : bool, optional
        If True the text on the title bar will be displayed in white, otherwise in black color (defaul is True)
    titleheight : int, optional
        Height of the title bar in pixels (default is 54)
    titleimageurl : str, optional
        String containing the url of the title bar backbround image (default is '')
    footercolor : str, optional
        Color to use fir the footer bar background (default is settings.color_second)
    footerdark : bool, optional
        If True the text on the footer bar will be displayed in white, otherwise in black color (defaul is False)
    footerheight : int, optional
        Height of the footer bar in pixels (default is 30)
    logoappurl : str, optional
        String containing the url of the application logo, to be displayed on the left side of the title bar (default is '')
    logowidth : int, optional
        Width in pixels of the application logo button (default is 40)
    on_logoapp : function, optional
        Python function to call when the user clicks on the pplication logo. The function will receive no parameters (default is None)
    copyrighttext : str, optional
        Text to display as copyright message on the footer bar (default is '')
    show_back : bool, optional
        If True a "back" button is displayed in the title bar (default is True)
    left_back : bool, optional
        If True the "back" button is displayed on the left side of the title bar (default is False)
    show_help : bool, optional
        If True a "help" button is displayed on the right side of the title bar (default is True)
    on_help : function, optional
        Python function to call when the user clicks the help button. The function will receive no parameters (default is None)
    logocreditsurl : str, optional
        String containing the url of the credits logo, to be displayed on the right side of the title bar (default is ''). If no url is passed, the logo of the European Commission is displayed
    show_credits : bool, optional
        If True a "credits" button is displayed on the right side of the title bar (default is True)
    on_credits : function, optional
        Python function to call when the user clicks the credits button. The function will receive no parameters (default is None)
    transition : str, optional
        Transition to be used for display and hide of the page (default is 'dialog-bottom-transition'). See: https://vuetifyjs.com/en/styles/transitions/ for a list of available transitions (substitute 'v-' with 'dialog-')
    persistent : bool, optional
        If True the page will be persistent and not close at the hitting of the "ESC" key (default is False)

    Examples
    --------
    Creation of an example page::
        
        from vois.vuetify import page
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()
        display(output)
        
        def onclose():
            pass

        def on_click():
            pass
            
        p = page.page('Application XYZ', 'Map page', output, onclose=onclose,
                      titlecolor='#008800', titledark=True,
                      footercolor='#cccccc', footerdark=False,
                      logoappurl='https://jeodpp.jrc.ec.europa.eu/services/shared/pngs/BDAP_Logo1024transparent.png',
                      on_logoapp=on_click, copyrighttext='European Commission - Joint Research Centre',
                      show_back=True, show_help=True, on_help=on_click,
                      show_credits=True, on_credits=on_click)
                      
        card = p.create()
        card.children = []
        p.open()


    .. figure:: figures/page.png
       :scale: 100 %
       :alt: page widget

       Example of a page
       

    """
    
    # Click on the APP logo
    def click_on_logoapp(self, *args):
        if not self.on_logoapp is None:
            self.on_logoapp()

    # Click on the back button
    def click_on_back(self, *args):
        try:
            self.close()
        except:
            pass
    
    # Click on the help button
    def click_on_help(self, *args):
        if not self.on_help is None:
            self.on_help()
    
    # Click on the credits button
    def click_on_credits(self, *args):
        if not self.on_credits is None:
            self.on_credits()


    # Internal method to create the childrens of the self.appbar instance
    def _create_appbar_children(self):
        textcolor = 'black'
        if self._titledark:
            textcolor = 'white'
        
        children = []
        
        if not self.logoapp is None:
            btn_logo = v.Btn(text=True, rounded=False, ripple=False, style_='width: %dpx; height: 40px;'%self._logowidth, class_='pa-0 ma-0 ml-1', children=[self.logoapp])
            btn_logo.on_event('click', self.click_on_logoapp)
            if len(self._appname) > 0:
                children.append(tooltip.tooltip("Info on %s"%self._appname, btn_logo))
            else:
                children.append(btn_logo)
            
        self.toolbarTitle = None
        if len(self._appname) > 0:
            self.toolbarTitle = v.ToolbarTitle(children=['%s:'%self._appname],  class_='pa-0 ma-0 mt-1 ml-1', style_='height: 30px; color: %s; font-size: 26;'%textcolor)
            children.append(self.toolbarTitle)
            
        self.toolbarSubtitle = None
        if len(self._title) > 0:
            self.toolbarSubtitle = v.ToolbarTitle(children=[self._title], class_='pa-0 ma-0 mt-1 ml-2', style_='height: 30px; color: %s; font-size: 26;'%textcolor)
            children.append(self.toolbarSubtitle)
        
        children.append(v.Spacer())
        
        self.buttons = []
        for cb in self.custom_buttons:
            iconname,tooltiptext,callback = cb
            btn = v.Btn(icon=True, class_="pa-0 ma-0 mt-1", dark=self._titledark, children=[v.Icon(children=[iconname])])
            btn.on_event('click', callback)
            self.buttons.append(btn)
            children.append(tooltip.tooltip(tooltiptext, btn))
        
        self.btn_back = None
        if self._show_back:
            self.btn_back = v.Btn(icon=True, class_="pa-0 ma-0 mt-1", dark=self._titledark, children=[v.Icon(children=['mdi-arrow-left'])])
            self.btn_back.on_event('click', self.click_on_back)
            elem = tooltip.tooltip("Close current page", self.btn_back)
            if self._left_back:
                children.insert(0, elem)
            else:
                children.append(elem)
        
        self.btn_help = None
        if self._show_help:
            self.btn_help = v.Btn(icon=True, class_="pa-0 ma-0 mt-1 mr-3", dark=self._titledark, children=[v.Icon(children=['mdi-help'])])
            self.btn_help.on_event('click', self.click_on_help)
            children.append(tooltip.tooltip("Display application help", self.btn_help))

        self.btn_credits = None
        if self._show_credits:
            self.btn_credits = v.Btn(text=True, rounded=False, ripple=False, style_='width: 170px;  height: %dpx;'%(self._titleheight-4), class_='pa-0 ma-0 mr-1', children=[self.logoCredits])
            self.btn_credits.on_event('click', self.click_on_credits)
            children.append(tooltip.tooltip("Open credits info", self.btn_credits))
        
        return children

    
    # Create the page and returns the card widget where the content of the page must be displayed
    def create(self):
        self.appbar = v.AppBar(height=self._titleheight, min_height=self._titleheight, max_height=self._titleheight, color=self._titlecolor,
                               children=self._create_appbar_children())

        if len(self._titleimageurl) > 0:
            self.appbar.src = self._titleimageurl


        # Content of the footer bar
        textcolor = 'black'
        if self._footerdark:
            textcolor = 'white'
            
        self.copyicon = v.Icon(class_="pa-0 ma-0 mr-2", small=True, color=textcolor, children=['mdi-copyright'])
        self.ctext = v.Card(flat=True, color=self._footercolor, style_='color: %s;'%textcolor, children=[self._copyrighttext])
        frow = v.Row(class_='pa-0 ma-0 mt-n2 mb-n3', justify="center", no_gutters=True, children=[self.copyicon,self.ctext])
        self.footer = v.Footer(color=self._footercolor, padless=True, children=[frow], class_='pa-0 ma-0', rounded=False,
                               height=self._footerheight, max_height=self._footerheight, min_height=self._footerheight,
                               style_='height: %dpx; overflow: hidden; border-bottom-left-radius: 0; border-bottom-right-radius: 0;'%self._footerheight)

        # Main content of the page: a card to be filled with custom content
        self.height = 'calc(100vh - %dpx)'%(self._titleheight+self._footerheight)
        self.card = v.Card(children=[], elevation=5,class_="pa-0 ma-0", style_='width: 100vw; max-width: 100vw; height: %s; max-height: %s;'%(self.height,self.height))
        return self.card
    
    
    # Open the dialog
    def open(self):
        self.dlg = dialogGeneric.dialogGeneric(title='', titleheight='0px', text='', show=True, no_click_animation=True,
                                               addclosebuttons=False, persistent=self.persistent, transition=self._transition,
                                               fullscreen=True, content=[self.appbar,self.card,self.footer], output=self.output)
        
        
    # Close the dialog
    def close(self):
        try:
            if not self.dlg is None:
                self.dlg.close()
                if not self.onclose is None:
                    self.onclose()
        except:
            pass
        
        self.dlg = None
        
    
    # Quick open and close of the page (to be called when the titleheight or the footerheight is changed)
    def refresh(self):
        if self.dlg is not None:
            self.close()
            oldtransition = self.transition
            self.transition = 'dialog-top-transition'
            self.open()
            self.transition = oldtransition
        
        
    # Add a custom buttom to the page (before the call to create!)
    def customButtonAdd(self, iconname, tooltiptext, callback):
        self.custom_buttons.append((iconname,tooltiptext,callback))   # Each item has an icon name, a tooltip string and a callback function
    
    # Remove all custom buttons
    def customButtonClear(self):
        self.custom_buttons = []
        
        
    #####################################################################################################################################################
    # Properties
    #####################################################################################################################################################
    @property
    def appname(self):
        """
        Get/Set the name of the application (displayed on the left side of the title bar)
        
        Returns
        --------
        name : str
            Name of the application

        Example
        -------
        Programmatically change the name of the application::
            
            p.appname = 'New app name'
            print(p.appname)
        
        """
        return self._appname
        
    @appname.setter
    def appname(self, name):
        self._appname = str(name)
        self.appbar.children = self._create_appbar_children()
        
        
    @property
    def title(self):
        """
        Get/Set the title of the page (displayed on the left side of the title bar)
        
        Returns
        --------
        title : str
            Title of the page

        Example
        -------
        Programmatically change the title of the page::
            
            p.title = 'New title'
            print(p.title)
        
        """
        return self._title
        
    @title.setter
    def title(self, title):
        self._title = str(title)
        self.appbar.children = self._create_appbar_children()
        
        
    @property
    def titlecolor(self):
        """
        Get/Set the color of the title bar
        
        Returns
        --------
        color : str
            Color of the title bar

        Example
        -------
        Programmatically change the color::
            
            p.titlecolor = 'red'
            print(p.titlecolor)
        
        """
        return self._titlecolor
        
    @titlecolor.setter
    def titlecolor(self, color):
        self._titlecolor = color
        self.appbar.color = self._titlecolor
        
        
    @property
    def titledark(self):
        """
        Get/Set the text color of the title bar
        
        Returns
        --------
        flag : bool
            If True the text on the title bar will be displayed in white, otherwise in black color

        Example
        -------
        Programmatically change the text color::
            
            p.titledark = True
            print(p.titledark)
        
        """
        return self._titledark
        
    @titledark.setter
    def titledark(self, flag):
        self._titledark = flag
        for btn in self.buttons:
            btn.dark = self._titledark

        if not self.btn_back is None:
            self.btn_back.dark = self._titledark
        if not self.btn_help is None:
            self.btn_help.dark = self._titledark
        
        textcolor = 'black'
        if self._titledark:
            textcolor = 'white'
            
        if not self.toolbarTitle is None:
            self.toolbarTitle.style_ = 'height: 30px; color: %s; font-size: 26;'%textcolor
        
        if not self.toolbarSubtitle is None:
            self.toolbarSubtitle.style_ = 'height: 30px; color: %s; font-size: 26;'%textcolor
            
            
    @property
    def titleheight(self):
        """
        Get/Set the height of the title bar
        
        Returns
        --------
        height : int
            Height of the title bar in pixels

        Example
        -------
        Programmatically change the title bar height::
            
            p.titleheight = 80
            print(p.titleheight)
        
        """
        return self._titleheight
        
    @titleheight.setter
    def titleheight(self, height):
        self._titleheight = height
        if self._titleheight < 20:  self._titleheight = 20
        if self._titleheight > 200: self._titleheight = 200
        self.appbar.height = self.appbar.min_height = self.appbar.max_height = self._titleheight
        
        if not self.btn_credits is None:
            self.btn_credits.style_ = 'width: 170px;  height: %dpx;'%(self._titleheight-4)
            
        self.height = 'calc(100vh - %dpx)'%(self._titleheight+self._footerheight)
        self.card.style_ = 'width: 100vw; max-width: 100vw; height: %s; max-height: %s;'%(self.height,self.height)
            
            
    @property
    def titleimageurl(self):
        """
        Get/Set the image to display on the title bar
        
        Returns
        --------
        imageurl : str
            String containing the url of the title bar backbround image

        Example
        -------
        Programmatically change the title bar background image::
            
            p.titleimageurl = 'https://.....'
            print(p.titleimageurl)
        
        """
        return self._titleimageurl
        
    @titleimageurl.setter
    def titleimageurl(self, imageurl):
        self._titleimageurl = imageurl

        if len(self._titleimageurl) > 0:
            self.appbar.src = self._titleimageurl
        else:
            self.appbar.src = None

            
    @property
    def footercolor(self):
        """
        Get/Set the color of the footer bar
        
        Returns
        --------
        color : str
            Color of the footer bar

        Example
        -------
        Programmatically change the color::
            
            p.footercolor = 'red'
            print(p.footercolor)
        
        """
        return self._footercolor
        
    @footercolor.setter
    def footercolor(self, color):
        self._footercolor = color
        self.ctext.color  = self._footercolor
        self.footer.color = self._footercolor
            
            
    @property
    def footerdark(self):
        """
        Get/Set the text color of the footer bar
        
        Returns
        --------
        flag : bool
            If True the text on the footer bar will be displayed in white, otherwise in black color

        Example
        -------
        Programmatically change the text color::
            
            p.footerdark = True
            print(p.footerdark)
        
        """
        return self._footerdark
        
    @footerdark.setter
    def footerdark(self, flag):
        self._footerdark = flag
        
        textcolor = 'black'
        if self._footerdark:
            textcolor = 'white'
            
        self.copyicon.color = textcolor
        self.ctext.style_   = 'color: %s;'%textcolor
            
            
    @property
    def footerheight(self):
        """
        Get/Set the height of the footer bar
        
        Returns
        --------
        height : int
            Height of the footer bar in pixels

        Example
        -------
        Programmatically change the footer bar height::
            
            p.footerheight = 20
            print(p.footerheight)
        
        """
        return self._footerheight
        
    @footerheight.setter
    def footerheight(self, height):
        self._footerheight = height
        if self._footerheight < 20:  self._footerheight = 16
        if self._footerheight > 200: self._footerheight = 200
        self.footer.height = self.footer.min_height = self.footer.max_height = self._footerheight
        self.footer.style_ = 'height: %dpx; overflow: hidden; border-bottom-left-radius: 0; border-bottom-right-radius: 0;'%self._footerheight
        
        self.height = 'calc(100vh - %dpx)'%(self._titleheight+self._footerheight)
        self.card.style_ = 'width: 100vw; max-width: 100vw; height: %s; max-height: %s;'%(self.height,self.height)

        
    @property
    def copyrighttext(self):
        """
        Get/Set the copyright text to display on the footer bar
        
        Returns
        --------
        text : str
            Text to display as copyright message on the footer bar

        Example
        -------
        Programmatically change the copyright text::
            
            p.copyrighttext = 'Founded by EU'
            print(p.copyrighttext)
        
        """
        return self._copyrighttext
        
    @copyrighttext.setter
    def copyrighttext(self, text):
        self._copyrighttext = text
        self.ctext.children = [self._copyrighttext]
        

    @property
    def show_back(self):
        """
        Show or hide the back button in the title bar
        
        Returns
        --------
        flag : bool
            If True a "back" button is displayed in the title bar

        Example
        -------
        Programmatically change the presence of the back button in the title bar::
            
            p.show_back = False
            print(p.show_back)
        
        """
        return self._show_back
        
    @show_back.setter
    def show_back(self, flag):
        self._show_back = flag
        self.appbar.children = self._create_appbar_children()
        
        
    @property
    def left_back(self):
        """
        Get/Set the flag that controls the position of the back button in the title bar
        
        Returns
        --------
        flag : bool
            If True the "back" button is displayed on the left side of the title bar

        Example
        -------
        Programmatically change the position of the back button in the title bar::
            
            p.left_back = False
            print(p.left_back)
        
        """
        return self._left_back
        
    @left_back.setter
    def left_back(self, flag):
        self._left_back = flag
        self.appbar.children = self._create_appbar_children()
        
        
    @property
    def show_help(self):
        """
        Show or hide the help button in the title bar
        
        Returns
        --------
        flag : bool
            If True a "help" button is displayed in the title bar

        Example
        -------
        Programmatically change the presence of the help button in the title bar::
            
            p.show_help = False
            print(p.show_help)
        
        """
        return self._show_help
        
    @show_help.setter
    def show_help(self, flag):
        self._show_help = flag
        self.appbar.children = self._create_appbar_children()

        
    @property
    def show_credits(self):
        """
        Show or hide the credits image in the title bar
        
        Returns
        --------
        flag : bool
            If True a "credits" button with an image is displayed in the title bar

        Example
        -------
        Programmatically change the presence of the credits button in the title bar::
            
            p.show_credits = False
            print(p.show_credits)
        
        """
        return self._show_credits
        
    @show_credits.setter
    def show_credits(self, flag):
        self._show_credits = flag
        self.appbar.children = self._create_appbar_children()
        
        
    @property
    def logoappurl(self):
        """
        Get/Set the image to display as the application logo on the left side of the title bar
        
        Returns
        --------
        imageurl : str
            String containing the url of application logo image

        Example
        -------
        Programmatically change the application logo in the title bar::
            
            p.logoappurl = 'data:image/png;base64,...'
            print(p.logoappurl)
        
        """
        return self._logoappurl
        
    @logoappurl.setter
    def logoappurl(self, imageurl):
        self._logoappurl = imageurl
        
        if len(self._logoappurl) > 0:
            self.logoapp = v.Img(class_='pa-0 ma-0 mr-2', max_width=self._logowidth, src=self._logoappurl)
        else:
            self.logoapp = None
        
        self.appbar.children = self._create_appbar_children()
        
        
    @property
    def logocreditsurl(self):
        """
        Get/Set the image to display as the application credits on the right side of the title bar
        
        Returns
        --------
        imageurl : str
            String containing the url of application credits image

        Example
        -------
        Programmatically change the application credits in the title bar::
            
            p.logocreditsurl = 'data:image/png;base64,...'
            print(p.logocreditsurl)
        
        """
        return self._logocreditsurl
        
    @logocreditsurl.setter
    def logocreditsurl(self, imageurl):
        self._logocreditsurl = imageurl
        
        if len(self._logocreditsurl) == 0:
            self._logocreditsurl = 'https://jeodpp.jrc.ec.europa.eu/services/shared/pngs/EC-JRC-logo_horizontal_EN_neg_transparent-background.png'

        self.logoCredits = v.Img(class_='pa-0 ma-0 mr-2', max_width=self._creditswidth, src=self._logocreditsurl)
        self.appbar.children = self._create_appbar_children()
        
               
    @property
    def logowidth(self):
        """
        Get/Set the width of the application logo button in the title bar
        
        Returns
        --------
        width : int
            Width of the application logo button in pixels

        Example
        -------
        Programmatically change the application logo button width::
            
            p.logowidth = 20
            print(p.logowidth)
        
        """
        return self._logowidth
        
    @logowidth.setter
    def logowidth(self, width):
        self._logowidth = width

        if len(self._logoappurl) > 0:
            self.logoapp = v.Img(class_='pa-0 ma-0 mr-2', max_width=self._logowidth, src=self._logoappurl)
        else:
            self.logoapp = None
        
        self.appbar.children = self._create_appbar_children()
        
        
    @property
    def creditswidth(self):
        """
        Get/Set the width of the application credits button in the title bar
        
        Returns
        --------
        width : int
            Width of the application credits button in pixels

        Example
        -------
        Programmatically change the application credits button width::
            
            p.creditswidth = 20
            print(p.creditswidth)
        
        """
        return self._creditswidth
        
    @creditswidth.setter
    def creditswidth(self, width):
        self._creditswidth = width
        self.logoCredits = v.Img(class_='pa-0 ma-0 mr-2', max_width=self._creditswidth, src=self._logocreditsurl)
        self.appbar.children = self._create_appbar_children()
        
        
    @property
    def transition(self):
        """
        Get/Set the transition mode to use for the opening of the page. Possible values are: 'dialog-bottom-transition', 'dialog-top-transition' or 'dialog-transition'
        
        Returns
        --------
        mode : str
            Transition mode, one of 'dialog-bottom-transition', 'dialog-top-transition' or 'dialog-transition'

        Example
        -------
        Programmatically change the transition mode::
            
            p.transition = 'dialog-transition'
            print(p.transition)
        
        """
        return self._transition
        
    @transition.setter
    def transition(self, mode):
        self._transition = mode
        
        
    @property
    def state(self):
        """
        Get/Set the multiple properties of the page
        
        Returns
        --------
        statusdict : dict
            Dictionary containing one or more properties

        Example
        -------
        Programmatically set some of the properties of a page::
            
            p.state = {'transition': 'dialog-top-transition', 'title': 'new title'}
            print(p.state)
        
        """
        return {x: getattr(self, x) for x in ['appname',
                                              'title',
                                              'titlecolor',
                                              'titledark',
                                              'titleheight',
                                              'titleimageurl',
                                              'footercolor',
                                              'footerdark',
                                              'footerheight',
                                              'copyrighttext',
                                              'show_back',
                                              'left_back',
                                              'show_help',
                                              'show_credits',
                                              'logoappurl',
                                              'logocreditsurl',
                                              'logowidth',
                                              'creditswidth',
                                              'transition'
                                             ]}
        
    @state.setter
    def state(self, statusdict):
        for key, value in statusdict.items():
            setattr(self, key, value)
            
        doRefresh = False
        if 'titleheight' in statusdict:
            if statusdict['titleheight'] != self.titleheight:
                doRefresh = True
            
        if 'footerheight' in statusdict:
            if statusdict['footerheight'] != self.footerheight:
                doRefresh = True
        
        if doRefresh:
            self.refresh()
        
        
    #####################################################################################################################################################
    # Initialization
    #####################################################################################################################################################
    def __init__(self,
                 appname,
                 title,
                 output,
                 onclose=None,
                 titlecolor=settings.color_first,
                 titledark=True,
                 titleheight=54,
                 titleimageurl='',
                 footercolor=settings.color_second,
                 footerdark=False,
                 footerheight=30,
                 logoappurl='',
                 logowidth=40,
                 on_logoapp=None,
                 copyrighttext='',
                 show_back=True,
                 left_back=False,
                 show_help=True,
                 on_help=None,
                 logocreditsurl='',
                 creditswidth=120,
                 show_credits=True,
                 on_credits=None,
                 transition='dialog-bottom-transition',
                 persistent=False):

        self.dlg = None
        
        self._appname = appname
        self._title   = title
        self.output   = output
        self.onclose  = onclose
        
        self._titlecolor    = titlecolor
        self._titledark     = titledark
        self._titleheight   = titleheight
        self._titleimageurl = titleimageurl
        
        self._footercolor  = footercolor
        self._footerdark   = footerdark
        self._footerheight = footerheight
        
        self._copyrighttext = copyrighttext
        self._show_back     = show_back
        self._left_back     = left_back
        self._show_help     = show_help
        self.on_help        = on_help
        self._show_credits  = show_credits
        self.on_credits     = on_credits
        self.on_logoapp     = on_logoapp
        
        self._transition    = transition
        self.persistent     = persistent
        
        self.custom_buttons = []   # Each item has an icon name, a tooltip string and a callback function
        
        self._logowidth      = logowidth
        self._creditswidth   = creditswidth
        
        self._logoappurl     = logoappurl
        self._logocreditsurl = logocreditsurl
        
        if len(self._logoappurl) > 0:
            self.logoapp = v.Img(class_='pa-0 ma-0 mr-2', max_width=self._logowidth, src=self._logoappurl)
        else:
            self.logoapp = None
        
        if len(self._logocreditsurl) == 0:
            self._logocreditsurl = 'https://jeodpp.jrc.ec.europa.eu/services/shared/pngs/EC-JRC-logo_horizontal_EN_neg_transparent-background.png'
        self.logoCredits = v.Img(class_='pa-0 ma-0 mr-2', max_width=self._creditswidth, src=self._logocreditsurl)



"""SVG heatmap chart from a pandas DataFrame."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import pandas as pd
from textwrap import wrap
from ipywidgets import HTML, widgets, Layout
from ipyevents import Event

try:
    from . import colors
    from .vuetify import fontsettings
except:
    import colors
    from vuetify import fontsettings

    
    
# Utility function to word-wrap a long text
def wordwrap(x, maxcharperline=80):
    return '<br>'.join(['<br>'.join(wrap(block, width=maxcharperline)) for block in x.splitlines()])



# Vertical text in SVG
def verticalText(text, x, y, size, color, fontfamily, weight):
    return '''
<text x="0" y="0" font-size="%f" fill="%s" font-weight="%d" style="font-family: %s; pointer-events: none;" transform="translate(%f,%f) rotate(-90)">%s</text>
''' % (size, color, weight, fontfamily, x, y, text)


###########################################################################################################################################################################
# Generic heatmap chart from a pandas dataframe containing only numbers (rows names are taken from the index). Returns an Output widget
###########################################################################################################################################################################
def heatmapChart(df,
                 width=100.0,             # width in vw coordinates
                 height=50.0,             # height in vh coordinates
                 hTitle=5.0,              # height of the Title bar
                 wTitle=15.0,             # width of the left space for row titles
                 columnTitleMaxChar=40,   # Max number of chars for the column names (longer names are split)
                 textRows='Row',
                 textColumns='Column',
                 textValues='Value',
                 title='Heatmap',
                 fontsize=1.0,
                 colorlist=['rgb(247,251,255)', 'rgb(222,235,247)', 'rgb(198,219,239)', 'rgb(158,202,225)', 'rgb(107,174,214)', 'rgb(66,146,198)', 'rgb(33,113,181)', 'rgb(8,81,156)', 'rgb(8,48,107)'],
                 textcolor='black',
                 textweight=400,
                 backcolor= 'white',
                 highlitecolor='#426bb4',
                 highliteback='#dddddd',
                 minvalue=0.0,
                 maxvalue=1.0,
                 decimals=2):
    """
    Creation of a heatmap chart given an input DataFrame containing only numbers. The index strings and column names are taken as names for rows and columns. The SVG chart has x coordinates expressed in vw coordinates and the y coordinates expressed invh coordinates. Rows and columns of the chart can be selected and the chart will be sorted on decreasing values (when a column is selected, the rows are sorted, and viceversa)
    
    Parameters
    ----------
    df : pandas DataFrame
        Input DataFrame containing only numbers
    width : float, optional
        Width of the chart in vw units (default is 100.0)
    height : float, optional
        Height of the chart in vh units (default is 50.0)
    hTitle : float, optional
        Height of the Title bar in percentage on the height of the chart (default is 5.0)
    wTitle : float, optional
        Width of the left space for row titles in percentage of the width pf the chart (default is 15.0)
    columnTitleMaxChar : int, optional
        Max number of chars for the column names (longer names are splitted), (default is 40)
    textRows : str, optional
        Title for the rows (default is 'Row')
    textColumns : str, optional
        Title for the columns (default is 'Column')
    textValues : str, optional
        Text to label the values in the tooltip when the mouse is over a cell of the chart (default is 'Value')
    title : str, optional
        Title of the chart (default is 'Heatmap')
    fontsize : float, optional
        Size of the standard font to use for values displayed in the X and Y axis in vh coordinates (default is 1.0vh). The chart title and the axis titles will be displayed with sizes proportional to the fontsize parameter (up to two times for the chart title)
    colorlist : list of colors, optional
        List of colors to assign to the country polygons (default is the Plotly px.colors.sequential.Blues, see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    textcolor : str, optional
        Color to use for rows and columns text (default is 'black')
    textweight : int, optional
        Weight of the text (default is 400). The chart title and the axis titles will be displayed with weight equal to textweight+100
    backcolor : str, optional
        Background color (default is 'white')
    highlitecolor : str, optional
        Color to use for displaying text of the selected row and colum (default is '#426bb4')
    highliteback : str, optional
        Color to use as background of the selected row and column (default is '#dddddd')
    minvalue : float, optional
        Minimum value of the DataFrame cells to be used for color assignment (default is 0.0)
    maxvalue : float, optional
        Minimum value of the DataFrame cells to be used for color assignment (default is 1.0)
    decimals : int, optional
        Number of decimals for the tooltip display of cell values (default is 2)
            
    Returns
    -------
        an instance of widgets.Output with the svg chart displayed in it

    Example
    -------
    Creation of a SVG heatmap chart to display a matrix of random numbers::

        from IPython.display import display
        import pandas as pd
        import numpy as np
        from vois import svgHeatmap
        
        df = pd.DataFrame(np.random.random((20,50)))
        display(svgHeatmap.heatmapChart(df, wTitle=7.0, hTitle=70)
        
    .. figure:: figures/heatmap.png
       :scale: 100 %
       :alt: svgHeatmap example

       Example of an interactive heatmap chart in SVG
        
    """
    
    # Dimension in SVG coordinates
    svgwidth = 100.0
    aspectratio = 0.5*height / width   # In landscape mode, usually the height is half the width dimension!!!
    svgheight = svgwidth * aspectratio

    if maxvalue <= minvalue: maxvalue = minvalue + 1.0
        
    # original df
    dfunsorted = df.copy()
    
    nrows = df.shape[0]
    hrow = (svgheight-hTitle)/nrows
    
    highlitecolumn = ''
    highliterow    = ''
    
    titlefontsize = fontsize*2
    
    #preserve = 'none'
    preserve = 'xMidYMid meet'
    
    # Calculates the SVG string
    def getSVG():
        svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" xml:space="preserve" viewBox="0 0 %f %f" preserveAspectRatio="%s" width="%fvw" height="%fvh">' % (svgwidth,svgheight, preserve, width,height)

        svg += '''
        <style type="text/css">
           @import url('%s');
            .cell:hover { cursor: pointer; stroke: #000044; stroke-width:0.1; stroke-dasharray:0.15,0.15; !important; }
        </style>
        ''' % (fontsettings.font_url)

        # Names of row titles in the first column
        rowTitles = list(df.index)
        y = hTitle
        for r in rowTitles:

            tcolor     = textcolor
            tweight    = textweight
            bcolor     = backcolor
            if r == highliterow:
                tcolor  = highlitecolor
                tweight = textweight + 100
                bcolor = highliteback
            svg += '<rect fill="%s" width="%f" height="%f" x="0.01" y="%f"></rect>' % (bcolor, wTitle*0.995, hrow*0.95, y+0.2)
            svg += '<text style="pointer-events: none" text-anchor="end" x="%f" y="%f" font-size="%f" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (wTitle-0.1, y+hrow/2.0+0.5, fontsize, 
                                                                                                                                                                        tcolor, fontsettings.font_name, tweight, r)
            y += hrow
            

        # Calculate color interpolator
        ci = colors.colorInterpolator(colorlist, minvalue, maxvalue)

        # Cells
        y = hTitle-0.01
        wcolumn = (svgwidth-wTitle) / len(df.columns)
        w = 0.9 * wcolumn

        wmod = w
        if wmod > 3.0: wmod = 3.0

        svg += '<rect fill="%s" width="%f" height="%f" x="0.01" y="%f" ></rect>' % (backcolor, wTitle-0.02, hTitle-0.01, 0.01)

        # Column names
        textangle = -90.0   # -45.0
        x = wTitle
        y = hTitle-0.01
        for c in df.columns:

            c = str(c)
            tcolor  = textcolor
            tweight = textweight
            bcolor  = backcolor

            if c == str(highlitecolumn):
                tcolor  = highlitecolor
                tweight = textweight + 100
                bcolor  = highliteback

            svg += '<rect fill="%s" width="%f" height="%f" x="%f" y="%f" ><title>%s</title></rect>' % (bcolor,w*1.05, hTitle-2.4, x, 2.5, c)

            if len(c) <= columnTitleMaxChar:
                svg += verticalText(c, x+w/2+0.2, y, fontsize, tcolor, fontsettings.font_name, tweight)
            else:
                text = wordwrap(c,columnTitleMaxChar)
                lines = text.split('<br>')

                if len(lines) == 2:
                    dx = wmod/3.5
                    cx = x + w/2
                    xx = cx - dx/3 + 0.35
                else:
                    dx = wmod/3.5
                    cx = x + w/2
                    xx = cx - 2*dx/3 + 0.15
                    
                for r in lines:
                    svg += verticalText(r, xx, y, fontsize, tcolor, fontsettings.font_name, tweight)
                    xx += dx


            x += wcolumn

        # Cells
        y = hTitle-0.01
        for r in df.index:
            x = wTitle
            for c in df.columns:
                value = df.at[r, c]
                color = ci.GetColor(value)
                svg += '<rect stroke-width="0.0" style="fill:%s;"  x="%f" width="%f" height="%f" y="%f"></rect>' % (color, x, w*1.2, hrow*1.2, y+0.25)

                x += wcolumn

            y += hrow


        # Cell highlights
        y = hTitle-0.01
        for r in df.index:
            sr = str(r)
            ptext = wordwrap(sr)

            x = wTitle
            for c in df.columns:
                sc = str(c)
                value = df.at[r, c]
                color = ci.GetColor(value)
                svalue = '{:.{prec}f}'.format(value, prec=decimals)
                svg += '<rect class="cell" stroke-width="0.0" style="fill:#ffffff00;"  x="%f" width="%f" height="%f" y="%f"><title>%s: %s\n%s: %s\n%s = %s</title></rect>' \
                       % (x, w*1.1, hrow, y+0.25, textRows, ptext, textColumns, sc, textValues, svalue)

                x += wcolumn

            y += hrow

            
        # Title texts
        svg += '<text style="pointer-events: none" text-anchor="end"    x="%f" y="%f" font-size="%f" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (wTitle-0.2, hTitle-0.1, fontsize*1.75, textcolor,
                                                                                                                                                                       fontsettings.font_name, textweight+100, textRows)
        svg += '<text style="pointer-events: none" text-anchor="end"    x="%f" y="%f" font-size="%f" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (wTitle-0.2, hTitle/2+fontsize, fontsize*1.75, textcolor,
                                                                                                                                                                       fontsettings.font_name, textweight+100, textColumns)
        svg += '<text dominant-baseline="middle"   text-anchor="middle" x="%f" y="%f" font-size="%f" fill="%s" style="font-family: %s;" font-weight="%d">%s</text>' % (svgwidth/2.0, 1.2, fontsize*2.0, textcolor,
                                                                                                                                                                       fontsettings.font_name, textweight+100, title)


        # Highlight of the big rect
        svg += '<rect class="cell" fill="#ffffff00" width="%f" height="%f" x="0.01" y="%f" ><title>Click to deselect rows and columns</title></rect>' % (wTitle-0.02, hTitle-0.01, 0.01)
        
        # Highlights of row titles in the first column
        y = hTitle
        for r in rowTitles:
            svg += '<rect class="cell" fill="#ffffff00" width="%f" height="%f" x="0.01" y="%f"><title>%s</title></rect>' % (wTitle*0.995, hrow*0.95, y+0.2, r)
            y += hrow
            
            
        # Highlight column names
        x = wTitle
        y = hTitle-0.01
        for c in df.columns:
            c = str(c)
            svg += '<rect class="cell" fill="#ffffff00" width="%f" height="%f" x="%f" y="%f" ><title>%s</title></rect>' % (w*1.05, hTitle-2.4, x, 2.5, c)
            x += wcolumn
            
        
        svg += '</svg>'
        return svg

    
    
    # Update the chart
    def updateChart():
        output.clear_output(wait=True)
        with output:
            display(HTML(getSVG()))
        
    
    #debug = widgets.Output()
    
    
    # Pixels to add to the output Widget in order to not see the scrollbars
    added_pixels_width  = 20
    added_pixels_height = 20
    
    
    # Management of 'click' event on the HEATMAP Output
    def handle_event_heatmap(event):
        nonlocal df, highlitecolumn, highliterow
        
        x = event['relativeX'] - added_pixels_width/2 - 2
        y = event['relativeY']
        w = event['boundingRectWidth']  - added_pixels_width
        h = event['boundingRectHeight'] - added_pixels_height
        ar = float(h) / float(w)
        
        # Calculate dimension and bounding box of chart in pixels
        if ar > aspectratio:
            chartw = w
            charth = w * aspectratio
            xmin   = 0
            xmax   = chartw
            ymin   = (h - charth)/2
            ymax   = ymin + charth
        else:
            charth = h
            chartw = h / aspectratio
            xmin   = (w - chartw)/2
            xmax   = xmin + chartw
            ymin   = 0
            ymax   = charth


        # Calculate position in percentage on the chart area in pixels
        if x >= xmin and x <= xmax and y >= ymin and y <= ymax:
            xp = 100.0 * (x-xmin) / float(chartw)
            yp = 100.0 * (y-ymin) / float(charth) - 1.0

            wTitlePercent = 100.0*wTitle/svgwidth
            hTitlePercent = 100.0*hTitle/svgheight

            nrows = df.shape[0]
            ncols = df.shape[1]
            
            wcol = (100.0 - wTitlePercent)/ncols
            hrow = (100.0 - hTitlePercent)/nrows
            
            #debug.clear_output()
            #with debug:
                #print('chartw=',chartw, '    charth=',charth, '      aspectratio=',aspectratio, '     ar=',ar)
                #print('xmin=',xmin,'   ymin=',ymin, '    xmax=',xmax,'   ymax=',ymax)
                #print('wcol=',wcol, '    hrow=',hrow)
                #print('xp=',xp,'    yp=',yp,'    wTitlePercent=',wTitlePercent, '    hTitlePercent=',hTitlePercent)
        
                
            # Click on the reset column
            if xp >= 0 and xp <= wTitlePercent and yp >= 0.0 and yp <= hTitlePercent:
                #with debug:
                #    print("RESET!")
                df = dfunsorted.copy()
                highlitecolumn = ''
                highliterow    = ''
                updateChart()
                
            # Click on a column
            elif yp >= 2.4 and yp <= hTitlePercent:
                if yp >= 8.0 and xp >= wTitlePercent:
                    #with debug:
                    #    print("COL!", int((xp - wTitlePercent) / wcol))
                    icol = int((xp - wTitlePercent) / wcol)
                    if icol >= 0 and icol < ncols:
                        highlitecolumn = df.columns[icol]
                        df = df.sort_values(highlitecolumn, ascending=False)  # Sort rows by descending values in one column
                        updateChart()
                        
            # Click on a row
            elif yp > hTitlePercent:
                if xp >= 0.0 and xp <= wTitlePercent:
                    #with debug:
                    #    print("ROW!", int((yp - hTitlePercent) / hrow))
                    irow = int((yp - hTitlePercent) / hrow)
                    if irow >= 0 and irow < nrows:
                        highliterow = df.index[irow]
                        df = df.sort_values(highliterow, axis=1, ascending=False)  # Sort columns by descending values in one row
                        updateChart()
                
    
    # Create the output widget
    output = widgets.Output(layout=Layout(width= 'calc(%fvw + %dpx)' % (width,added_pixels_width), 
                                          height='calc(%fvh + %dpx)' % (height,added_pixels_height),
                                          margin='0px 0px 0px 0px')) #border='1px dashed green'))
    
    # Create the Event manager
    dh = Event(source=output, watched_events=['click'])
    dh.on_dom_event(handle_event_heatmap)

    updateChart()
        
    return output #, debug



"""Upload local image: dialog box with image preview"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2024
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Imports
from ipywidgets import widgets, Layout
import ipyvuetify as v

# Vois imports
from vois import cssUtils, colors
from vois.vuetify import settings, upload, dialogGeneric, dialogWait

# Python imports
from PIL import Image


#####################################################################################################################################################
# Upload local image: dialog box with image preview
#####################################################################################################################################################
class UploadImage():
    
    # Initialization
    def __init__(self,
                 output,
                 message='Click to select image to upload',
                 label='Image:',
                 accept='image/png, image/jpeg, image/bmp',
                 title='Select an image',
                 color=settings.color_first,
                 dark=settings.dark_mode,
                 titleheight=40,
                 width=620,
                 onOK=None,                                       # Called passing the url string of the selected image or the empty string if no image is selected
                 onCancel=None):                                  # Called with no argument
        
        self.output       = output
        self.title        = title
        self.color        = color
        self.dark         = dark
        self.titleheight  = titleheight
        self.width        = width
        self.onOK         = onOK
        self.onCancel     = onCancel
        
        cssUtils.allSettings(self.output)
        
        self.image    = None
        self.imageurl = ''
        
        self.wait = None
        
        self.preview = widgets.Output(layout=Layout(height='600px'))
        self.u = upload.upload(accept=accept, label=label, onchanging=self.onFileSelected, onchange=self.onFileUpload, placeholder=message, multiple=False)

        spacerY = v.Html(tag='div', style_='width: 0px; height: 20px;')
        self.content = v.Card(flat=True, class_='pa-0, ma-0 mt-n4 ml-4 mr-4', children=[widgets.VBox([self.u.draw(), spacerY, self.preview])])
        
        
    # Open the dialog-box
    def show(self):
        self.preview.clear_output()
        self.u.clear()
        dialogGeneric.dialogGeneric(title=self.title, on_ok=self._internal_onOK, on_cancel=self._internal_onCancel, on_close=self.onCancel,
                                    text='   ', color=self.color, dark=self.dark, titleheight=self.titleheight, width=self.width,
                                    show=True, addclosebuttons=True, addokcancelbuttons=True,
                                    fullscreen=False, content=[self.content], output=self.output)
        
    # Just after the file is selected
    def onFileSelected(self):
        self.wait = dialogWait.dialogWait(text='Uploading file...', output=self.output, color=self.color, dark=self.dark)
        

    # Selection of an image to upload
    def onFileUpload(self, files):
        # If at least one file has been selected      
        if len(files) > 0:
            self.preview.clear_output()
            for f in files:
                self.image = Image.open(f['file_obj'])
                self.imageurl = colors.image2Base64(self.image)
                with self.preview:
                    display(v.Img(src=self.imageurl))

        # No files selected
        else:
            self.image    = None
            self.imageurl = ''
            self.preview.clear_output()
    
        if self.wait is not None:
            self.wait.close()
    
    
    # Selection done
    def _internal_onOK(self):
        if self.onOK is not None:
            self.onOK(self.imageurl)
    
    # Exit without selection
    def _internal_onCancel(self):
        if self.onCancel is not None:
            self.onCancel()



"""Utility functions to update the URL of the page that launched the dashboard."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display, HTML
            
# Update the visualized URL in the browser
def urlUpdate(url, output):
    """
    Update the URL visualized in the top bar of the browser.
    
    Parameters
    ----------
    url : str
        Partial url to add to the current browser's page key/values

    Example
    -------
    Add a key/value pair to the current browser URL::

        from vois import urlUpdate
        from ipywidgets import widgets, Layout
        from IPython.display import display

        output = widgets.Output(layout=Layout(width='0px', height='0px'))
        display(output)
        
        urlUpdate.urlUpdate('?test=3', output)
    
    Note
    ----
        If the dashboard is created using the :class:`app.app` class, it is preferable to use the function :func:`app.app.urlUpdate` that doesn't need the output parameter and uses the Output widget created inside the app instance itself (and invisible to the users)
    
    """
    js = "<script>window.history.replaceState({ additionalInformation: 'Updated the URL with JS' }, '', '%s');</script>" % url
    with output:
            display(HTML(js))



"""Add tooltip text to a widget: returns a "modified" widget to be used instead of the original one."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v

try:
    from . import settings
except:
    import settings


#####################################################################################################################################################
# Add tooltip text to a widget: returns a "modified" widget to be used instead of the original one
#####################################################################################################################################################
def tooltip(text, widget):
    """
    Add a tooltip to a widget.

    Parameters
    ----------
    text : str
        Text of the tooltip
    widget: ipyvuetify widget
        Instance of the widget to which the tooltip has to be added
        
    Returns
    -------
    v.Item
        An ipyvuetify v.Item widget having a v.Tooltip as its only child

    Example
    -------
    Add a tooltip to a switch widget::
    
        from vois.vuetify import tooltip, switch
        from IPython.display import display

        s = switch.switch(True, "Activate the notification")
        t = tooltip.tooltip('Select to activate the notification of events to the user', s.draw())
        display(t)
        
    .. figure:: figures/tooltip.png
       :scale: 100 %
       :alt: tooltip widget

       Tooltip added to a switch widget.
    """
    if len(text) > 0:
        widget.v_on = 'tooltip.on'
        item = v.Item(class_="pa-0 ma-0", 
                      children=[ v.Tooltip(color=settings.tooltip_backcolor, transition="scale-transition", bottom=True, left=True,
                                           v_slots=[{'name': 'activator', 'variable': 'tooltip', 'children': widget }],
                                           children=[text]) ])
        return item
    else:
        #return widget
        
        # Tooltip present but not activated
        #widget.v_on = 'tooltip.on'
        item = v.Item(class_="pa-0 ma-0", 
                      children=[ v.Tooltip(color=settings.tooltip_backcolor, transition="scale-transition", bottom=True, left=True,
                                           v_slots=[{'name': 'activator', 'variable': 'tooltip', 'children': widget }],
                                           children=[text]) ])
        return item




"""Widget for the creation and editing of color palettes."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
from ipywidgets import widgets, Layout
from IPython.display import display, HTML
import random
import base64
import json

try:
    from . import settings
    from . import sortableList
    from . import palettePicker
    from . import ColorPicker
    from . import switch
    from . import tooltip
    from . import dialogGeneric
    from . import dialogWait
    from . import dialogMessage
    from . import selectSingle
    from . import upload
except:
    import settings
    import sortableList
    import palettePicker
    import ColorPicker
    import switch
    import tooltip
    import dialogGeneric
    import dialogWait
    import dialogMessage
    import selectSingle
    import upload

    
    
# Utility to return items from a list of colors
def colorlist2Items(colors=[]):
    """
    Utility function to convert a list of colors to a list of items to pass as parameter to the :py:class:`paletteEditor.paletteEditor` class
    
    Parameters
    ----------
    colorlist : list of str, optional
        List of string representing colors in the format '#RRGGBB' (default is [])
        
    Return
    ------
        A list of items ready to be passed to the :py:class:`paletteEditor.paletteEditor` class constructor
        
    """
    return [{ "value": 0,  "class": "", "color": x } for x in colors]
    
    

# Utility: 3 integers to '#RRGGBB'
def RGB(r,g,b):
    return '#{:02X}{:02X}{:02X}'.format(r, g, b)
    

    
# Class paletteEditor
class paletteEditor():
    """
    Widget for the creation and editing of color palettes.
        
    Parameters
    ----------
    title : str, optional
        Title of the palette (default is '')
    items : list of dicts, optional
        List of dicts containing "value", "class" and "color" (default is [])
    interpolate : bool, optional
        Flag to control the interpolation of the colorlist: if True the palette will be displayed by adding intermediate colors (default is True)
    width : int, optional
        Width of the widget in pixels (default is 420)
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    dark : bool, optional
        Flag to invert the text and backcolor (default is the value of settings.dark_mode)
    onchange : function, optional
        Python function to call when the user changes the order of colors or removes a color. The function will receive no parameters as input (default is None)
    buttonstooltip : bool, optional
        If True, the buttons to mode, add, remove colors and assign values to colors will have a tooltip (default is True)

    Example
    -------
    Example of a widget to create and edit a color palette::
        
        from vois.vuetify import paletteEditor
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def onchange():
            with output:
                print('onchange!')

        p = paletteEditor(items=[], width=450, onchange=onchange)

        display(p.draw())
        display(output)

    .. figure:: figures/paletteEditor.png
       :scale: 100 %
       :alt: card widget

       Example of a widget to create and edit a color palette
    """
    
    # Initialization
    def __init__(self, title='', items=[], interpolate=True, width=420, maxheightlist=600,
                 color=settings.color_first, dark=settings.dark_mode,
                 onchange=None, buttonstooltip=True):

        self.width          = width
        self.color          = color
        self.dark           = dark
        self.onchange       = onchange
        self.buttonstooltip = buttonstooltip
        
        self.outputdialog  = widgets.Output(layout=Layout(width='0px', min_width='0px', height='0px'))
        self.outputtoolbar = widgets.Output()
        self.outputtitle   = widgets.Output()
        self.outputpalette = widgets.Output()
        self.outputlist    = widgets.Output()
        self.outputpreview = widgets.Output(layout=Layout(width='480px', min_width='480px', height='50px'))
        

        self.tfTitle = v.TextField(v_model=title, label="Title", color=self.color, class_="pa-0 ma-0 mt-3", style_="width: %dpx; max-width: %dpx"%(self.width,self.width))

        self.sp = v.Html(tag='div', class_="pa-0 ma-0 mr-3", children=[''])

        self.palette_text = ''
        
        self.tf = v.TextField(v_model="Palette", label="Palette file name", autofocus=True, 
                              color=self.color, class_="pa-0 ma-0 ml-4 mr-4")
        
        custompalettes = [
            { "name": "Simple", "colors": ['#000000', '#FF0000', '#00FF00', '#0000FF', '#FFFF00', '#FF00FF', '#00FFFF', '#FFFFFF']},

            { "name": "Dem",    "colors": [RGB(255,255,170), RGB( 39,168, 39), RGB( 11,128, 64), RGB(255,255,  0), RGB(255,186,  3),
                                           RGB(158, 30,  2), RGB(110, 40, 10), RGB(138, 94, 66), RGB(255,255,255)]},

            { "name": "NDVI",   "colors": [RGB(120,69,25), RGB(255,178,74), RGB(255,237,166), RGB(173,232,94),
                                           RGB(135,181,64), RGB(3,156,0), RGB(1,100,0), RGB(1,80,0)]}
        ]

        families = ['carto', 'cmocean', 'cyclical', 'diverging', 'plotlyjs', 'qualitative', 'sequential', 'custom']
        family   = 'sequential'

        self.selfam = selectSingle.selectSingle('Family:', families, selection=family, width=160, onchange=self.onchangeFamily, marginy=1, clearable=False)
        self.pp = palettePicker.palettePicker(family=family, custompalettes=custompalettes, label='Palette:', height=26)

        self.upload_file = upload.upload(accept="application/json", multiple=False, show_progress=False, onchange=self.on_upload_palette,
                                         label='Palette file:', placeholder='Click to select the palette to upload', width='480px', margins="pa-0 ma-0 ml-4")
        
        # Top buttons
        self.new  = v.Btn(icon=True, class_="mr-0", dark=self.dark, children=[v.Icon(children=['mdi-file-outline'])])
        self.new.on_event('click', self.onnew)

        self.upload  = v.Btn(icon=True, class_="mr-0", dark=self.dark, children=[v.Icon(children=['mdi-folder-open-outline'])])
        self.upload.on_event('click', self.onupload)

        self.download  = v.Btn(icon=True, class_="mr-0", dark=self.dark, children=[v.Icon(children=['mdi-content-save'])])
        self.download.on_event('click', self.ondownload)

        self.select  = v.Btn(icon=True, class_="mr-0", dark=self.dark, children=[v.Icon(children=['mdi-folder-search-outline'])])
        self.select.on_event('click', self.onselect)
        
        self.revert  = v.Btn(icon=True, class_="mr-0", dark=self.dark, children=[v.Icon(children=['mdi-autorenew'])])
        self.revert.on_event('click', self.onrevert)

        self.random  = v.Btn(icon=True, class_="mr-0", dark=self.dark, children=[v.Icon(children=['mdi-palette'])])
        self.random.on_event('click', self.onrandom)
        
        with self.outputtoolbar:
            display(v.Row(no_gutters=True, style_="min-width: 300px;", justify="start", class_="pa-0 ma-0",
                          children=[tooltip.tooltip("New palette",self.new),
                                    tooltip.tooltip("Load a palette from local disk",self.upload),
                                    tooltip.tooltip("Save current palette to local disk",self.download),
                                    tooltip.tooltip("Select one of the pre-defined palettes",self.select),
                                    tooltip.tooltip("Revert palette order",self.revert),
                                    tooltip.tooltip("Assign random colors",self.random)]))
 
        with self.outputtitle:
            display(self.tfTitle)
            
        # Bottom buttons
        self.zero  = v.Btn(icon=True, class_="mr-0", dark=self.dark, children=[v.Icon(children=['mdi-numeric-0'])])
        self.zero.on_event('click', self.onzero)
        
        self.asc   = v.Btn(icon=True, class_="mr-0", dark=self.dark, children=[v.Icon(children=['mdi-numeric-1'])])
        self.asc.on_event('click', self.onasc)
        
        self.desc  = v.Btn(icon=True, class_="mr-0", dark=self.dark, children=[v.Icon(children=['mdi-numeric-negative-1'])])
        self.desc.on_event('click', self.ondesc)
        
        self.sw = switch.switch(interpolate, "Interpolate", color=self.color, onchange=self.__internal_onchange)

        self.s = sortableList.sortableList(items=items, width=self.width, maxheightlist=maxheightlist, outlined=False, dark=self.dark,
                                           allowNew=True, itemNew=self.itemNew, itemContent=self.itemContent,
                                           bottomContent=[#self.sp,self.sp,
                                                          tooltip.tooltip("Set all values to 0",self.zero),
                                                          tooltip.tooltip("Set increasing values starting from first color",self.asc),
                                                          tooltip.tooltip("Set decreasing values starting from last color", self.desc),
                                                          tooltip.tooltip("Interpolate colors",self.sw.draw())],
                                           onchange=self.__internal_onchange,
                                           buttonstooltip=self.buttonstooltip,
                                           tooltipadd='Add new color', tooltipdown='Move color down',
                                           tooltipup='Move color up', tooltipremove='Remove color')

        self.updatePalette()
        
        with self.outputlist:
            display(self.s.draw())

            
    # Change family in the palette picker
    def onchangeFamily(self):
        family = self.selfam.value
        if family == 'carto' or family == 'qualitative':
            interpolate = False
        else:
            interpolate = True
        self.pp.updatePalettes(family,interpolate)
        
        
    # New palette
    def onnew(self, widget, event, data):
        self.s.items         = []
        self.sw.value        = True
        self.tfTitle.v_model = ''


    # Select
    def onselect(self, widget, event, data):
        def on_ok():
            dlg = dialogWait.dialogWait(text='Loading palette...', output=self.outputdialog)
            self.s.items = colorlist2Items(self.pp.colors)
            self.sw.value = self.pp.interpolate
            dlg.close()
            
        content1 = v.Row(no_gutters=True, justify="start", class_="pa-0 ma-0 ml-6", children=[self.selfam.draw()])
        content2 = v.Row(no_gutters=True, justify="start", class_="pa-0 ma-0 ml-6", children=[self.pp.draw()])

        dlg = dialogGeneric.dialogGeneric(title='Palette selection', text='Please select one of the predefined palettes:',
                                          show=True, addclosebuttons=False, width=600,
                                          addokcancelbuttons=True, on_ok=on_ok,
                                          fullscreen=False, content=[widgets.VBox([content1, content2])], output=self.outputdialog)

        
    # Called when a .json file is selected for upload
    def on_upload_palette(self, files):
        self.outputpreview.clear_output(wait=False)
        if len(files) > 0:
            f = files[0]
            self.palette_text = f['file_obj'].read().decode("utf-8")

            try:
                j = json.loads(self.palette_text)
                if ("format" in j) and ("interpolate" in j) and ("items" in j) and (j["format"] == "BDAP palette 1.0"):
                    items = j["items"]
                    interpolate = j["interpolate"]
                    colors = [x["color"] for x in items]
                    with self.outputpreview:
                        display(v.Card(outlined=True, dark=self.dark, class_="pa-0 ma-0", style_='width: 400px; max-width: 400px; height: 40px; max-height: 40px;' ,
                                children=[v.Img(class_="pa-0 ma-1", src=palettePicker.image2Base64(palettePicker.paletteImage(colors, width=360, height=29, interpolate=interpolate)))]))
                else:
                    self.upload_file.clear()
                    e = dialogMessage.dialogMessage(title='Error',
                                                    text='The uploaded file is not in the \"BDAP palette 1.0\" format',
                                                    addclosebuttons=False, show=True, width=400, output=self.outputdialog)
            except:
                self.upload_file.clear()
                e = dialogMessage.dialogMessage(title='Error',
                                                text='Cannot read the uploaded file!',
                                                addclosebuttons=False, show=True, width=400, output=self.outputdialog)
        else:
            self.palette_text = ''

    
    # Called when the file upload dialog is closed with the "OK" button
    def on_upload_ok(self):
        if len(self.palette_text) > 2:
            dlg = dialogWait.dialogWait(text='Loading palette...', output=self.outputdialog)
            try:
                j = json.loads(self.palette_text)
                if ("format" in j) and ("interpolate" in j) and ("items" in j) and (j["format"] == "BDAP palette 1.0"):
                    self.s.items, self.sw.value = j["items"], j["interpolate"]
                    if "title" in j: self.tfTitle.v_model = j["title"]
                    else:            self.tfTitle.v_model = ''
                else:
                    e = dialogMessage.dialogMessage(title='Error',
                                                    text='The uploaded file is not in the \"BDAP palette 1.0\" format',
                                                    addclosebuttons=False, show=True, width=400, output=self.outputdialog)
            except:
                e = dialogMessage.dialogMessage(title='Error',
                                                text='Cannot read the uploaded file!',
                                                addclosebuttons=False, show=True, width=400, output=self.outputdialog)
                
            dlg.close()
            
        self.palette_text = ''
        self.upload_file.clear()
            
    
    # Called when the file upload dialog is closed with the "cancel" button
    def on_upload_cancel(self):
        self.palette_text = ''
        self.upload_file.clear()
        
        
    # Upload
    def onupload(self, widget, event, data):
        self.outputpreview.clear_output()
        dlg = dialogGeneric.dialogGeneric(title='Load a palette from local disk', text='', show=True, 
                                          addclosebuttons=False, width=520,
                                          addokcancelbuttons=True, on_ok=self.on_upload_ok, on_cancel=self.on_upload_cancel,
                                          fullscreen=False, content=[self.upload_file.draw(), self.outputpreview], output=self.outputdialog)

    

    # Direct download of a .txt file containing a string
    def downloadText(self, textobj, fileName="palette.json"):
        string_bytes  = textobj.encode("ascii","ignore")
        base64_bytes  = base64.b64encode(string_bytes)
        base64_string = base64_bytes.decode("ascii")

        self.outputdialog.clear_output()
        with self.outputdialog:
            display(HTML('<script>function downloadURI(uri, name) { var link = document.createElement("a"); link.download = name; link.href = uri; link.click();} downloadURI("data:application/octet-stream;charset=utf-8;base64,' + base64_string + '","' + fileName + '"); </script>'))

            
        # Download
    def ondownload(self, widget, event, data):

        def on_ok():
            items = self.s.items
            j = { "format":      "BDAP palette 1.0",
                  "interpolate": self.sw.value,
                  "items"      : items,
                  "title"      : self.tfTitle.v_model
                }
            txt = json.dumps(j)
            filename = self.tf.v_model
            if filename[-5:] != ".json": filename += ".json"
            self.downloadText(txt, fileName=filename)
        
        dlg = dialogGeneric.dialogGeneric(title='Save current palette to local disk', text='', show=True, 
                                          addclosebuttons=False, width=500,
                                          addokcancelbuttons=True, on_ok=on_ok,
                                          fullscreen=False, content=[self.tf], output=self.outputdialog)


    # Revert palette order
    def onrevert(self, widget, event, data):
        dlg = dialogWait.dialogWait(text='Updating palette...', output=self.outputdialog)
        self.s.items = list(reversed(self.s.items))
        dlg.close()


    # Assign random colors
    def onrandom(self, widget, event, data):
        dlg = dialogWait.dialogWait(text='Updating palette...', output=self.outputdialog)
        items = self.s.items
        for item in items:
            item['color'] = "#" + ''.join([random.choice('0123456789ABCDEF') for j in range(6)])
        self.s.items = items
        dlg.close()
        
    
    
    # Set all values to zero
    def onzero(self, widget, event, data):
        dlg = dialogWait.dialogWait(text='Updating palette...', output=self.outputdialog)
        items = self.s.items
        for item in items:
            item['value'] = 0
        self.s.items = items
        dlg.close()
            
        
    # Set all increasing values starting from first color
    def onasc(self, widget, event, data):
        dlg = dialogWait.dialogWait(text='Updating palette...', output=self.outputdialog)
        items = self.s.items
        if len(items) > 0:
            start = items[0]['value']
            for index, item in enumerate(items):
                item['value'] = start + index
            self.s.items = items
        dlg.close()

    
    # Set all decreasing values starting from last color
    def ondesc(self, widget, event, data):
        dlg = dialogWait.dialogWait(text='Updating palette...', output=self.outputdialog)
        items = self.s.items
        if len(items) > 0:
            start = items[-1]['value']
            for index, item in enumerate(items):
                item['value'] = start - len(items) + 1 + index
            self.s.items = items
        dlg.close()

    
    # Update the palette image
    def updatePalette(self):
        colors = []
        for item in self.s.items:
            colors.append(item['color'])
            
        self.outputpalette.clear_output(wait=True)
        with self.outputpalette:
            display(v.Card(outlined=True, dark=self.dark, style_='width: %dpx; max-width: %dpx; height: 40px; max-height: 40px;' %(self.width,self.width),
                    children=[v.Img(class_="pa-0 ma-1", src=palettePicker.image2Base64(palettePicker.paletteImage(colors, width=self.width-20, 
                                                                                                                  height=29, interpolate=self.sw.value)))]))
            
        
    # Creation of a new item
    def itemNew(self):
        return { "value": 0,  "class": "", "color": "#FF0000" }


    # Content of an item
    def itemContent(self, item, index):

        def onvalue(widget, event, data):
            item["value"] = int(data)

        def onclass(widget, event, data):
            item["class"] = data

        def oncolor():
            item["color"] = cp.color
            self.updatePalette()

        netw = self.width - 140
        tfvalue = v.TextField(label='Value:', value=item['value'], color=self.color, type="number", dense=True, style_="max-width: %dpx"%(int(0.25*netw)), class_="pa-0 ma-0 mt-2")
        tfvalue.on_event('input', onvalue)

        tfclass = v.TextField(label='Class:', value=item['class'], color=self.color, dense=True, style_="max-width: %dpx"%(int(0.74*netw)), class_="pa-0 ma-0 mt-2")
        tfclass.on_event('input', onclass)

        cp = colorPicker.colorPicker(color=item['color'], dark=self.dark, width=30, show_swatches=False, onchange=oncolor)

        sp = v.Html(tag='div', class_="pa-0 ma-0 mr-3", children=[''])

        return [ v.Row(class_="pa-0 ma-0 ml-2", no_gutters=True, children=[tfvalue, sp, tfclass, sp, cp.draw()]) ]


    # Manage onchange on the sortableList widget
    def __internal_onchange(self, arg=None):
        self.updatePalette()
        if self.onchange:
            self.onchange()


    # Returns the vuetify object to display
    def draw(self):
        """Returns the ipyvuetify object to display (a v.Html object displaying two output widgets)"""
        return v.Html(tag='div',children=[widgets.VBox([v.Html(tag='div',children=[''], class_='pa-0 ma-0 mb-2'),
                                                        widgets.HBox([self.outputtoolbar,self.outputdialog]),self.outputtitle,self.outputpalette,self.outputlist])])

    
    # colors property
    @property
    def colors(self):
        """
        Get/set the colors of the palette.
        
        Returns
        --------
        colorlist : list of strings in '#RRGGBB' format
            List of colors of the palette

        Example
        -------
        Get the edited palette colors::
            
            print(editor.colors)
        
        """
        return [x['color'] for x in self.s.items]


    # Set the colors
    @colors.setter
    def colors(self, colorlist):
        dlg = dialogWait.dialogWait(text='Updating palette...', output=self.outputdialog)
        items = self.s.items
        for index,item in enumerate(items):
            item['color'] = colorlist[index%len(colorlist)]
        self.s.items = items
        dlg.close()
        
    
    # items property
    @property
    def items(self):
        """
        Get/set the items of the palette.
        
        Returns
        --------
        items : list of dicts
            List of dicts containing "value", "class" and "color"

        Example
        -------
        Print the edited palette items::
            
            print(editor.items)
        
        """
        return self.s.items

    
    # Set the items
    @items.setter
    def items(self, items):
        dlg = dialogWait.dialogWait(text='Updating palette...', output=self.outputdialog)
        self.s.items = items
        dlg.close()
        
        
    # title property
    @property
    def title(self):
        """
        Get/set the title of the palette.
        
        Returns
        --------
        t : str
            Title of the palette
        """
        return self.tfTitle.v_model

    
    # Set the title
    @title.setter
    def title(self, t):
        self.tfTitle.v_model = str(t)

        
    # interpolate property
    @property
    def interpolate(self):
        """
        Get/set the interpolate flag.
        
        Returns
        --------
        flag : bool
            Interpolate flag
        """
        return self.sw.value

    
    # Set the interpolate
    @interpolate.setter
    def interpolate(self, flag):
        self.sw.value = flag
        self.updatePalette()
        


"""Generic modal dialog-box to ask input from the user."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from IPython.display import display
import ipyvuetify as v

try:
    from . import tooltip
    from . import settings
except:
    import tooltip
    import settings
    

#####################################################################################################################################################
# Display a generic dialog with a close button (and possibly fullscreen)
#####################################################################################################################################################
class dialogGeneric():
    """
    Generic modal dialog-box to ask input from the user.
        
    Parameters
    ----------
    title : str, optional
        Title of the dialog-box to be displayed in the top toolbar (default is '')
    text : str, optional
        Text to display on top of the dialog-box body (default is '')
    color: str, optional
        Color of the title bar of the dialog (default is settings.color_first)
    dark : bool, optional
        Flag that controls the color of the text in foreground (if True, the text will be displayed in white, elsewhere in black)
    show : bool, optional
        Flag to immediately show the dialog-box upon creation (default is False)
    content: list of ipyvuetify widgets, optional
        List of ipyvuietify widgets to be displayed in the body of the dialog-box (default is [])
    width : int or str, optional
        Width of the dialog-box. If an integer is passed the width is intended in pixels. Default is 500 pixels
    fullscreen : bool, optional
        If True, the dialog-box is opened in fullscreen mode (default is False)
    persistent : bool, optional
        If True, clicking outside of the dialog or pressing esc key will not deactivate it (default is False)
    no_click_animation : bool, optional
        If True, disables the bounce effect when clicking outside of a dialog's content when using the persistent property (default is False)
    addclosebuttons : bool, optional
        If True, the dialog will have a 'close' button in the top toolbar (default is True)
    addokcancelbuttons : bool, optional
        If True, the dialog will have 'ok' and 'cancel' buttons in the bottom row (default is False)
    on_ok : function, optional
        Python function to call when the user clicks on the OK button. The function will receive no parameters (default is None)
    on_cancel : function, optional
        Python function to call when the user clicks on the CANCEL button. The function will receive no parameters (default is None)
    on_close : function, optional
        Python function to call when the user clicks on the CLOSE button. The function will receive no parameters (default is None)
    transition : str, optional
        Transition to use for the dialog display and close (default is 'dialog-fade-transition'. See: https://vuetifyjs.com/en/styles/transitions/ for a list of available transitions (substitute 'v-' with 'dialog-'')
    output : ipywidgets.Output, optional
        Output widget on which the widget has to be displayed
    titleheight : str, optional
        Height of the title toolbar. It can be: 'prominent', 'dense', 'extended' or a value in pixels (default is 'dense')
    custom_icon : str, optional
        Name of the optional icon to display in the top toolbar (default is '')
    custom_tooltip : str, optional
        Tooltip to display when hovering on the custom icon in the top toolbar (default is '')
    custom_icon_onclick : function, optional
        Python function to call when the user clicks on the custom icon on the topbar. The function will receive no parameters (default is None)
            
    Example
    -------
    Creation and display of a modal dialog-box containing a switch widget::
        
        from vois.vuetify import dialogGeneric, switch
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()
        display(output)

        def on_s_change(value):
            with output:
                print(value)

        s = switch.switch(True, 'PNG format', onchange=on_s_change)

        dlg = dialogGeneric.dialogGeneric(title='Settings',
                                          text='Please select the format for download:',
                                          show=True, addclosebuttons=True, width=600, 
                                          fullscreen=False, content=[s.draw()], output=output)
                                  
    .. figure:: figures/dialogGeneric.png
       :scale: 100 %
       :alt: dialogGeneric widget

       Example of a dialogGeneric containing a switch widget.
   """
        
    def __init__(self, title='', text='', color=settings.color_first, dark=settings.dark_mode, show=False, content=[], width=500, fullscreen=False,
                 persistent=False, no_click_animation=False, 
                 addclosebuttons=True, addokcancelbuttons=False, on_ok=None, on_cancel=None, on_close=None,
                 transition='dialog-fade-transition', output=None, titleheight="dense", customclass="",
                 custom_icon='', custom_tooltip='', custom_icon_onclick=None):
        
        self.on_ok     = on_ok
        self.on_cancel = on_cancel
        self.on_close  = on_close
        
        self.custom_icon_onclick = custom_icon_onclick
        
        text = text.replace('<br>','\n')
        vvv = text.split('\n')

        if dark:
            buttontext = settings.textcolor_dark
            styletext = 'color: ' + settings.textcolor_dark + ';'
        else:
            buttontext = settings.textcolor_notdark
            styletext = 'color: ' + settings.textcolor_notdark + ';'

        # Close button
        if addclosebuttons:
            bclose = v.Btn(text=True, children=['Close'], color=buttontext)
            bclose.on_event('click', self.close)
            bclose = tooltip.tooltip('Close the dialog',bclose)
        else:
            bclose = ''

        # Close X button
        if addclosebuttons:
            bx = v.Btn(icon=True, children=[v.Icon(children=['mdi-close'])], color=buttontext)
            bx.on_event('click', self.close)
            bx = tooltip.tooltip('Close the dialog',bx)
        else:
            bx = ''

        # Custom icon on the top toolbar
        if len(custom_icon) > 0:
            bi = v.Btn(icon=True, children=[v.Icon(children=[custom_icon])], color=buttontext)
            bi.on_event('click', self.oncustomicon)
            if len(custom_tooltip) > 0:
                bi = tooltip.tooltip(custom_tooltip,bi)
        else:
            bi = ''
            
            
        # OK and Cancel buttons
        if addokcancelbuttons:
            self.bok = v.Btn(text=True, children=['OK'])
            self.bok.on_event('click', self.__internal_on_ok)

            self.bcancel = v.Btn(text=True, children=['Cancel'])
            self.bcancel.on_event('click', self.__internal_on_cancel)

            r = [v.Row(no_gutters=True, justify="end", class_="pa-0 ma-0 mt-4", children=[self.bok,self.bcancel])]
        else:
            self.bok     = None
            self.bcancel = None
            r = []
            
        # Toolbar
        ttitle  = v.ToolbarTitle(children=[title], style_=styletext)
        spacer  = v.Spacer()
        titems  = v.ToolbarItems(children=[bi,bclose])
        toolbar = v.Toolbar(height=titleheight, dark=dark, color=color, children=[bx,ttitle,spacer,titems])
        
        # Dialog
        clist = [toolbar]
        if len(text) > 0:
            clist += [ v.CardText(class_='pa-0 ma-0 mt-9', children=['']) ]
            clist += [ v.CardText(children=[x], class_="mt-n5") for x in vvv ]
        clist += content + r
        
        if type(width)==int:
            swidth = "%dpx"%width
        else:
            swidth = str(width)
        self.dialog = v.Dialog(width=swidth, v_model=show, fullscreen=fullscreen, transition=transition,
                               persistent=persistent, no_click_animation=no_click_animation,
                               style_="background-color: transparent; z-index:20001;", 
                               content_class=customclass,
                               children=[v.Card(children=clist)])
        
        # Display of the dialog
        if not output is None:
            with output:
                display(self.dialog)
                
        
    # Open the dialog
    def show(self):
        """Open the dialog."""
        self.dialog.v_model = True

    # Close the dialog
    def close(self, *args):
        """Close the dialog."""
        self.dialog.v_model = False
        if not self.on_close is None:
            self.on_close()

    # Click on the custom icon
    def oncustomicon(self, *args):
        if not self.custom_icon_onclick is None:
            self.custom_icon_onclick()
        

    # Cliked ok
    def __internal_on_ok(self,*args):
        self.dialog.v_model = False
        if not self.on_ok is None:
            self.on_ok()

    # Cliked cancel
    def __internal_on_cancel(self,*args):
        self.dialog.v_model = False
        if not self.on_cancel is None:
            self.on_cancel()

    # Enable/disable the ok button
    @property
    def okdisabled(self):
        """
        Get/Set the disabled status of the ok button.
        
        Returns
        --------
        flag : bool
            True if the ok button is disabled, False otherwise

        Example
        -------
        Programmatically set the disabled status and print it::
            
            dlg.okdisabled = True
            print(dlg.okdisabled)
        
        """
        if self.bok is None:
            return True
        else:
            return self.bok.disabled
    
    
    # Set the disabled status of the ok button
    @okdisabled.setter
    def okdisabled(self, flag):
        if not self.bok is None:
            self.bok.disabled = bool(flag)
        